// Ansteuerung eines HD44780 kompatiblen LCD im 4-Bit-Interfacemodus
// http://www.mikrocontroller.net/articles/AVR-GCC-Tutorial
//
// Die Pinbelegung ist ber defines in lcd-routines.h einstellbar
//
// 18.3.2009, Trax: Auswahl belibiger Ports bei der initialisierung

#include <avr/io.h>
#include "lcd.h"
#include <util/delay.h>
#include <stdlib.h>
#include "../Screens.h"
#include "../Ports.h"


tLCDPins* g_Pins = 0;
extern tGpio g_Gpio[];

// LCD Befehle
#define CLEAR_DISPLAY 0x01
#define CURSOR_HOME   0x02

// erzeugt den Enable-Puls
void lcd_enable(void)
{
   // Bei Problemen ggf. Pause gem Datenblatt des LCD Controllers einfgen
   // http://www.mikrocontroller.net/topic/81974#685882
   if(g_Pins)
	   *(g_Gpio[g_Pins[EN].Port].pPort)  |= (1<<g_Gpio[g_Pins[EN].Port].nBit);
   else
	   LCD_PORT |= (1<<LCD_EN);
   _delay_us(1);
   // Bei Problemen ggf. Pause gem Datenblatt des LCD Controllers verlngern
   // http://www.mikrocontroller.net/topic/80900
   if(g_Pins)
	   *(g_Gpio[g_Pins[EN].Port].pPort) &= ~(1<<g_Gpio[g_Pins[EN].Port].nBit);
   else
	   LCD_PORT &= ~(1<<LCD_EN);
}

// setzt das register bit
void lcd_cmd (uint8_t cmd)
{
	if(cmd)
	{
		if(g_Pins)
			*(g_Gpio[g_Pins[RS].Port].pPort) &= ~(1<<g_Gpio[g_Pins[RS].Port].nBit);    // RS auf 0 setzen
		else
			LCD_PORT &= ~(1<<LCD_RS);
	}
	else
	{
		if(g_Pins)
			*(g_Gpio[g_Pins[RS].Port].pPort)  |= (1<<g_Gpio[g_Pins[RS].Port].nBit);     // RS auf 1 setzen
		else
			LCD_PORT |= (1<<LCD_RS);
	}
}

// sendet ein nibble
void lcd_nibble(unsigned char temp)
{
   temp = temp & 0x0F;
   if(g_Pins)
   {
	   *(g_Gpio[g_Pins[D4].Port].pPort) &= ~(1<<g_Gpio[g_Pins[D4].Port].nBit);
	   if(temp & 0x01)
		   *(g_Gpio[g_Pins[D4].Port].pPort)  |= (1<<g_Gpio[g_Pins[D4].Port].nBit);

	   *(g_Gpio[g_Pins[D5].Port].pPort) &= ~(1<<g_Gpio[g_Pins[D5].Port].nBit);
	   if(temp & 0x02)
		   *(g_Gpio[g_Pins[D5].Port].pPort)  |= (1<<g_Gpio[g_Pins[D5].Port].nBit);

	   *(g_Gpio[g_Pins[D6].Port].pPort) &= ~(1<<g_Gpio[g_Pins[D6].Port].nBit);
	   if(temp & 0x04)
		   *(g_Gpio[g_Pins[D6].Port].pPort)  |= (1<<g_Gpio[g_Pins[D6].Port].nBit);

	   *(g_Gpio[g_Pins[D7].Port].pPort) &= ~(1<<g_Gpio[g_Pins[D7].Port].nBit);
	   if(temp & 0x08)
		   *(g_Gpio[g_Pins[D7].Port].pPort)  |= (1<<g_Gpio[g_Pins[D7].Port].nBit);
   }
   else
   {
	   LCD_PORT &= 0x0F;
	   LCD_PORT |= temp << 4;
   }

   lcd_enable();
}

// sendet ein byte
void lcd_byte(unsigned char temp, uint8_t cmd)
{
   lcd_cmd(cmd);

   // oberes Nibble holen
   lcd_nibble(temp >> 4);               // setzen

   // unteres Nibble holen und maskieren
   lcd_nibble(temp);				// setzen

   _delay_us(42);
}

// Initialisierung:
// Muss ganz am Anfang des Programms aufgerufen werden.
void lcd_init(void* Pins)
{
   g_Pins = Pins;

   if(Pins)
   {
	   int i;
	   for(i=0;i<6;i++)
		   *(g_Gpio[g_Pins[i].Port].pDdr) |= (1<<g_Gpio[g_Pins[i].Port].nBit);
   }
   else
   {
	   LCD_DDR |= LCD_DB | (1<<LCD_RS) | (1<<LCD_EN);   // Port auf Ausgang schalten
   }

   // muss 3mal hintereinander gesendet werden zur Initialisierung
   _delay_ms(15);
   lcd_cmd(1);
   lcd_nibble(0x30);
   _delay_ms(5);
   lcd_enable();
   _delay_ms(5);
   lcd_enable();
   _delay_ms(5);

   // 4 Bit Modus aktivieren
   lcd_nibble(0x20);
   _delay_ms(1);

   // 4Bit / 2 Zeilen / 5x7
   lcd_byte(0x28, 1);

   // Display ein / Cursor aus / kein Blinken
   lcd_byte(0x0C, 1);

   // inkrement / kein Scrollen
   lcd_byte(0x06, 1);

   lcd_clear();
}

// Sendet den Befehl zur Lschung des Displays
void lcd_clear(void)
{
	lcd_byte(CLEAR_DISPLAY, 1);
   _delay_ms(5);
}

// Sendet den Befehl: Cursor Home
void lcd_home(void)
{
	lcd_byte(CURSOR_HOME, 1);
   _delay_ms(5);
}

// setzt den Cursor in Zeile y (1..4) Spalte x (0..15)
void set_cursor(uint8_t x, uint8_t y)
{
  uint8_t tmp;

  switch (y) {
    case 1: tmp=0x80+0x00+x; break;    // 1. Zeile
    case 2: tmp=0x80+0x40+x; break;    // 2. Zeile
    case 3: tmp=0x80+0x10+x; break;    // 3. Zeile
    case 4: tmp=0x80+0x50+x; break;    // 4. Zeile
    default: return;                   // fr den Fall einer falschen Zeile
  }
  lcd_byte(tmp, 1);
}

// Schreibt einen String auf das LCD
void lcd_string(const char *data)
{
    while(*data) {
    	lcd_byte(*data, 0);
        data++;
    }
}
