/*
 * Copyright (c) 2003, Adam Dunkels.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * This file is part of the uIP TCP/IP stack
 *
 * $Id: telnetd.c,v 1.2 2006/06/07 09:43:54 adam Exp $
 *
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "../uip/uip.h"
#include "telnetd.h"
#include "../../shell.h"

#define ISO_nl       0x0a
#define ISO_cr       0x0d
#define ISO_bs       0x08 // putty on serial: 0x7F

#define TELNET_IAC   255
#define TELNET_WILL  251
#define TELNET_WONT  252
#define TELNET_DO    253
#define TELNET_DONT  254

/*---------------------------------------------------------------------------*/
void shell_quit(struct telnetd_state* s) {
	s->state = STATE_CLOSE;
}
/*---------------------------------------------------------------------------*/
static void sendline(char *line, struct telnetd_state* s)
{
	uint8_t i;
	for (i = 0; i < TELNETD_CONF_NUMLINES; ++i) {
		if (s->lines[i] == NULL) {
			s->lines[i] = line;
			break;
		}
	}
	if (i == TELNETD_CONF_NUMLINES) {
		free(line);
	}
}
/*---------------------------------------------------------------------------*/
void shell_output(struct telnetd_state* s, int Return, const prog_char * _Format, ...)
{
	if(!s)
		return;

	va_list argptr;
	va_start(argptr, _Format);
	int len = vsnprintf_P(s->buf, sizeof(s->buf), _Format, argptr);
	va_end(argptr);

	if(Return)
	{
		if (len > TELNETD_CONF_LINELEN) {
			len = TELNETD_CONF_LINELEN;
		}
		s->buf[len] = ISO_cr;
		s->buf[len + 1] = ISO_nl;
		s->buf[len + 2] = 0;
		len+=2;
	}

	char *line = malloc(len+1);
	if (line != NULL)
	{
		memcpy(line, s->buf, len+1);

		//petsciiconv_toascii(line, TELNETD_CONF_LINELEN);
		sendline(line, s);
	}
	s->bufptr = 0;
}
/*---------------------------------------------------------------------------*/
void telnetd_init(void) {
	uip_listen(HTONS(PORT_TELNETD));
	//shell_init();
}
/*---------------------------------------------------------------------------*/
void acked(struct telnetd_state* s) {
	static unsigned int i;

	while (s->numsent > 0) {
		free(s->lines[0]);
		for (i = 1; i < TELNETD_CONF_NUMLINES; ++i) {
			s->lines[i - 1] = s->lines[i];
		}
		s->lines[TELNETD_CONF_NUMLINES - 1] = NULL;
		--s->numsent;
	}
}
/*---------------------------------------------------------------------------*/
static void senddata(struct telnetd_state* s) {
	static char *bufptr, *lineptr;
	static int buflen, linelen;

	bufptr = uip_appdata;
	buflen = 0;
	for (s->numsent = 0; s->numsent < TELNETD_CONF_NUMLINES && s->lines[s->numsent]
			!= NULL; ++s->numsent) {
		lineptr = s->lines[s->numsent];
		linelen = strlen(lineptr);
		//if (linelen > TELNETD_CONF_LINELEN) {
		//	linelen = TELNETD_CONF_LINELEN;
		//}
		if (buflen + linelen < uip_mss()) {
			memcpy(bufptr, lineptr, linelen);
			bufptr += linelen;
			buflen += linelen;
		} else {
			break;
		}
	}
	uip_send(uip_appdata, buflen);
}
/*---------------------------------------------------------------------------*/
static void closed(struct telnetd_state* s) {
	static unsigned int i;

	for (i = 0; i < TELNETD_CONF_NUMLINES; ++i) {
		if (s->lines[i] != NULL) {
			free(s->lines[i]);
		}
	}
}
/*---------------------------------------------------------------------------*/
static void get_char(u8_t c, struct telnetd_state* s) {
	if (c == ISO_nl) {
		return;
	}

	if(c == ISO_bs)
	{
		if(s->bufptr > 0)
			--s->bufptr;
		return;
	}

	s->buf[(int) s->bufptr] = c;
	if (s->buf[(int) s->bufptr] == ISO_cr || s->bufptr == sizeof(s->buf) - 1) {
		s->buf[(int) s->bufptr] = 0;			/*      petsciiconv_topetscii(s.buf, TELNETD_CONF_LINELEN);*/
		shell_input(s->buf, s);
		s->bufptr = 0;
	} else {
		++s->bufptr;
	}
}
/*---------------------------------------------------------------------------*/
static void sendopt(u8_t option, u8_t value, struct telnetd_state* s) {
	char *line = malloc(4);
	if (line != NULL)
	{
		line[0] = TELNET_IAC;
		line[1] = option;
		line[2] = value;
		line[3] = 0;
		sendline(line, s);
	}
}
/*---------------------------------------------------------------------------*/
static void newdata(char *dataptr, u16_t len, struct telnetd_state* s) {
	u8_t c;

	while (len > 0 && s->bufptr < sizeof(s->buf)) {
		c = *dataptr;
		++dataptr;
		--len;
		switch (s->state) {
		case STATE_IAC:
			if (c == TELNET_IAC) {
				get_char(c,s);
				s->state = STATE_NORMAL;
			} else {
				switch (c) {
				case TELNET_WILL:
					s->state = STATE_WILL;
					break;
				case TELNET_WONT:
					s->state = STATE_WONT;
					break;
				case TELNET_DO:
					s->state = STATE_DO;
					break;
				case TELNET_DONT:
					s->state = STATE_DONT;
					break;
				default:
					s->state = STATE_NORMAL;
					break;
				}
			}
			break;
		case STATE_WILL:
			/* Reply with a DONT */
			sendopt(TELNET_DONT, c, s);
			s->state = STATE_NORMAL;
			break;

		case STATE_WONT:
			/* Reply with a DONT */
			sendopt(TELNET_DONT, c, s);
			s->state = STATE_NORMAL;
			break;
		case STATE_DO:
			/* Reply with a WONT */
			sendopt(TELNET_WONT, c, s);
			s->state = STATE_NORMAL;
			break;
		case STATE_DONT:
			/* Reply with a WONT */
			sendopt(TELNET_WONT, c, s);
			s->state = STATE_NORMAL;
			break;
		case STATE_NORMAL:
			if (c == TELNET_IAC) {
				s->state = STATE_IAC;
			} else {
				get_char(c,s);
			}
			break;
		}
	}
}
void newchar(char Char, struct telnetd_state* s)
{
	if(s->state == STATE_CLOSE)
		s->state = STATE_NORMAL;
	newdata(&Char,1,s);
}
/*---------------------------------------------------------------------------*/
void telnetd_appcall(struct telnetd_state* s) {
	static unsigned int i;
	if (uip_connected()) {
		/*    tcp_markconn(uip_conn, &s);*/
		for (i = 0; i < TELNETD_CONF_NUMLINES; ++i) {
			s->lines[i] = NULL;
		}
		s->bufptr = 0;
		s->state = STATE_NORMAL;

		shell_start(s);
	}

	if (s->state == STATE_CLOSE) {
		s->state = STATE_NORMAL;
		uip_close();
		return;
	}

	if (uip_closed()||
	uip_aborted() ||
	uip_timedout()) {
		closed(s);
	}

	if(uip_acked()) {
		acked(s);
	}

	if(uip_newdata()) {
		newdata((char *) uip_appdata, uip_datalen(), s);
	}

	if(uip_rexmit() ||
			uip_newdata() ||
			uip_acked() ||
			uip_connected() ||
			uip_poll()) {
		senddata(s);
	}
}
/*---------------------------------------------------------------------------*/
